
template <typename elem, int order>
class flat {
    static_assert (order >= 0, "multicomplex order must be nonnegative");
    static_assert (order < 25, "sanity_check -- modify if desired");

    vector<elem> guts;

    static constexpr int raw_size {1 << order};

    template <int x_order>
    typename std::enable_if <(x_order > 0), void>::type
    help_lower (multicomplex<elem,x_order> const & mc) {
        help_lower<x_order-1> (mc.real);
        help_lower<x_order-1> (mc.imag);
    }

    template <int x_order>
    typename std::enable_if <(x_order == 0), void>::type
    help_lower (multicomplex<elem,0> const & mc) {
        guts.push_back (mc.sole);
    }

    template <int x_order> static
    typename std::enable_if <(x_order > 0), multicomplex<elem,x_order>>::type
    help_raise (vector<elem> const & g, int const lo, int const hi) {
        int const md {lo + (hi - lo) / 2};
        return multicomplex<elem,x_order> {
            help_raise<x_order-1> (g, lo, md),
            help_raise<x_order-1> (g, md, hi)
        };
    }

    template <int x_order> static
    typename std::enable_if <(x_order == 0), multicomplex<elem,x_order>>::type
    help_raise (vector<elem> const & g, int const lo, int const hi) {
        return multicomplex<elem,x_order> {g.at(lo)};
    }

    void view_i (ostream & o, int const lo, int const hi, int const dp) const {
        if ((lo + 1) == hi) {
            elem const e {guts.at(lo)};
            if (e >= 0)
                o << "+ " << +e;
            else
                o << "- " << -e;
            return;
        }
        int const md {lo + (hi - lo) / 2};
        o << "(";
        view_i (o, lo, md, dp - 1);

        o << " ";
        if (dp > 1) o << "+ ";

        view_i (o, md, hi, dp - 1);
        o << "*i" << dp << ")";
    }

    void view_k (ostream & o, int const lo, int const hi) const {
        if ((lo + 1) == hi) {
            o << guts.at(lo);
            return;
        }
        int const md {lo + (hi - lo) / 2};
        o << "[";
        view_k (o, lo, md);
        o << " ";
        view_k (o, md, hi);
        o << "]";
    }

public:
    flat () = delete;
    flat (flat const &) = default;
    flat (flat &&) = default;
    flat & operator= (flat const &) = default;
    flat & operator= (flat &&) = default;
    ~flat () = default;

    flat (multicomplex<elem,order> const & mc) {
        guts.reserve (raw_size);
        help_lower<order> (mc);
    }

    flat (vector<elem> const & v) : guts {v} {
        if (guts.size() != raw_size)
            throw fail {"flat::flat ... input vector is wrong size"};
    }

    elem const & at (int const j) const
        { return guts.at(j); }

    elem & at (int const j)
        { return guts.at(j); }

    int size () const
        { return raw_size; }

    explicit operator multicomplex<elem,order> () const
        { return help_raise<order> (guts, 0, raw_size); }

    explicit operator vector<elem> const & () const { return guts; }

    void view_i (ostream & o) const {
        view_i (o, 0, raw_size, order);
    }
    void view_j (ostream & o) const {
        o << "(";

        for (int j {0}; j < raw_size; ++j) {
            if (j > 0) o << " ";
            elem const & e {guts.at(j)};
            if (e >= 0)
                o << "+ " << +e;
            else
                o << "- " << -e;
            o << "*j" << j;
        }

        o << ")";
    }

    void view_k (ostream & o) const {
        view_k (o, 0, raw_size);
    }
};

template <typename elem, int order>
ostream & operator<< (ostream & o, flat<elem, order> const & f) {
    f.view_k (o); return o;
}

