
/* Your compiler may want these declarations:

template <typename elem, int order>
class multicomplex;

template <typename elem>
class multicomplex<elem, 0>;

*/

template <typename elem, int order>
class multicomplex {
    static_assert (order >= 0, "multicomplex order must be nonnegative");
    static_assert (order < 25, "sanity_check -- modify if desired");

    static constexpr int off {1 << (order - 1)};
    void view_j (ostream &, int const) const;
public:
    using elem_type = elem;
    static constexpr int order_value {order};
    using one_less = multicomplex<elem, order-1>;

    one_less real, imag; // these are the only data

/// constructors and related items

    multicomplex (
        one_less const & r,
        one_less const & i
    //  You might prefer these defaults:
    //  one_less const & r = one_less::zero,
    //  one_less const & i = one_less::zero
    ) : real {r}, imag {i} { }

    multicomplex () = delete;
    multicomplex (multicomplex const &) = default;
    multicomplex (multicomplex &&) = default;
    multicomplex & operator= (multicomplex const &) = default;
    multicomplex & operator= (multicomplex &&) = default;
    ~multicomplex () = default;

    static multicomplex const zero;
    static multicomplex const real_unit;
    static multicomplex const imag_unit;

    // WDTCRT means "why does the compiler requrire this?"
    template <typename WDTCRT = void>
    static typename std::enable_if <
        std::is_same<elem,double>::value
        && std::is_void<WDTCRT>::value
        , multicomplex>::type
    random () {
        return multicomplex {
            one_less::random(),
            one_less::random()
        };
    }

/// unary operators

    multicomplex operator+ () const {
        return multicomplex {+ real, + imag};
    }
    multicomplex operator~ () const {
        return multicomplex {+ real, - imag};
    }
    multicomplex operator- () const {
        return multicomplex {- real, - imag};
    }

/// addition

    template <int o_order>
    typename std::enable_if <(order > o_order), multicomplex<elem,order>>::type
    operator+ (multicomplex<elem, o_order> const & o) const {
        return multicomplex<elem, order> {real + o, imag};
    }

    template <int o_order>
    typename std::enable_if <(order < o_order), multicomplex<elem,o_order>>::type
    operator+ (multicomplex<elem, o_order> const & o) const {
        return multicomplex<elem, o_order> {*this + o.real, + o.imag};
    }

    template <int o_order>
    typename std::enable_if <(order == o_order), multicomplex<elem,order>>::type
    operator+ (multicomplex<elem, o_order> const & o) const {
        return multicomplex<elem, order> {real + o.real, imag + o.imag};
    }

/// subtraction

    template <int o_order>
    typename std::enable_if <(order > o_order), multicomplex<elem,order>>::type
    operator- (multicomplex<elem, o_order> const & o) const {
        return multicomplex<elem, order> {real - o, imag};
    }

    template <int o_order>
    typename std::enable_if <(order < o_order), multicomplex<elem,o_order>>::type
    operator- (multicomplex<elem, o_order> const & o) const {
        return multicomplex<elem, o_order> {*this - o.real, - o.imag};
    }

    template <int o_order>
    typename std::enable_if <(order == o_order), multicomplex<elem,order>>::type
    operator- (multicomplex<elem, o_order> const & o) const {
        return multicomplex<elem, order> {real - o.real, imag - o.imag};
    }

/// multiplication

    template <int o_order>
    typename std::enable_if <(order > o_order), multicomplex<elem,order>>::type
    operator* (multicomplex<elem, o_order> const & o) const {
        return multicomplex<elem, order> {real * o, imag * o};
    }

    template <int o_order>
    typename std::enable_if <(order < o_order), multicomplex<elem,o_order>>::type
    operator* (multicomplex<elem, o_order> const & o) const {
        return multicomplex<elem, o_order> {*this * o.real, *this * o.imag};
    }

    template <int o_order>
    typename std::enable_if <(order == o_order), multicomplex<elem,order>>::type
    operator* (multicomplex<elem, o_order> const & o) const {
        return multicomplex<elem, order> {
            real * o.real - imag * o.imag,
            real * o.imag + imag * o.real
        };
    }

    multicomplex operator* (elem const & o) const {
        return multicomplex {real * o, imag * o};
    }

/// division

    template <int o_order>
    typename std::enable_if <(order >= o_order), multicomplex<elem,order>>::type
    operator/ (multicomplex<elem,o_order> const & o) const {
        multicomplex const & numer {*this};
        multicomplex<elem,o_order  > const & denom {o};
        multicomplex<elem,  order  > const new_numer {numer * ~denom};
        multicomplex<elem,o_order-1> const new_denom {denom.fold()};
        return new_numer / new_denom;
    }

    template <int o_order>
    typename std::enable_if <(order < o_order), multicomplex<elem,o_order>>::type
    operator/ (multicomplex<elem,o_order> const & o) const {
        multicomplex const & numer {*this};
        multicomplex<elem,o_order  > const & denom {o};
        multicomplex<elem,o_order  > const new_numer {numer * ~denom};
        multicomplex<elem,o_order-1> const new_denom {denom.fold()};
        return new_numer / new_denom;
    }

    multicomplex operator/ (multicomplex<elem,0> const & o) const {
        multicomplex const & numer {*this};
        multicomplex<elem,0> const & denom {o};
        return numer / denom.sole;
    }

    multicomplex operator/ (elem const & o) const {
        return multicomplex {real / o, imag / o};
    }

    one_less fold () const {
        return real * real + imag * imag;
    }

/// selection by position

private:
    template <int x_order>
    static typename
        std::enable_if <(order > x_order), multicomplex<elem,x_order> const &>::type
    priv_at_con (multicomplex<elem,order> const & source, int const n) {
        static constexpr int half {1 << (order - x_order - 1)};
        if (n < half)
            return one_less::template priv_at_con<x_order> (source.real, n);
        else
            return one_less::template priv_at_con<x_order> (source.imag, n - half);
    }

    template <int x_order>
    static typename
        std::enable_if <(order > x_order), multicomplex<elem,x_order> &>::type
    priv_at_var (multicomplex<elem,order> & source, int const n) {
        static constexpr int half {1 << (order - x_order - 1)};
        if (n < half)
            return one_less::template priv_at_var<x_order> (source.real, n);
        else
            return one_less::template priv_at_var<x_order> (source.imag, n - half);
    }

    template <int x_order>
    static typename
        std::enable_if <(order == x_order), multicomplex<elem,x_order> &>::type
    priv_at_var (multicomplex<elem,order> & source, int const n) {
        return source;
    }

    template <int x_order>
    static typename
        std::enable_if <(order == x_order), multicomplex<elem,x_order> const &>::type
    priv_at_con (multicomplex<elem,order> const & source, int const n) {
        return source;
    }

    // INTENTIONALLY OMITTED:
    //
    // template <int x_order>
    // static typename
    //     std::enable_if <(order < x_order), multicomplex<elem,x_order> const &>::type
    // priv_at_con (multicomplex<elem,order> const & source, int const n);
    //
    // template <int x_order>
    // static typename
    //     std::enable_if <(order < x_order), multicomplex<elem,x_order> &>::type
    // priv_at_var (multicomplex<elem,order> & source, int const n);
public:
    template <int x_order>
    multicomplex<elem,x_order> const & at_con (int const n) const {
        static_assert (x_order >= 0, "x_order < 0");
        static_assert (x_order <= order, "x_order > order");
        static int const full {1 << (order - x_order)};
        if (n < 0) throw fail {"multicomplex::at_con ... n < 0"};
        if (n >= full) throw fail {"multicomplex::at_con ... n >= full"};
        return priv_at_con<x_order> (*this, n);
    }

    template <int x_order>
    multicomplex<elem,x_order> & at_var (int const n) {
        static_assert (x_order >= 0, "x_order < 0");
        static_assert (x_order <= order, "x_order > order");
        static int const full {1 << (order - x_order)};
        if (n < 0) throw fail {"multicomplex::at_var ... n < 0"};
        if (n >= full) throw fail {"multicomplex::at_var ... n >= full"};
        return priv_at_var<x_order> (*this, n);
    }

/// miscellaneous

    elem norm () const {
        return real.norm() + imag.norm();
    }

    void view_i (ostream &) const;
    void view_j (ostream &) const;
    void view_k (ostream &) const;

    template<typename elem2, int order2>
    friend class multicomplex;
};

template <typename elem, int order>
ostream & operator<< (ostream & o, multicomplex<elem, order> const & mc) {
    mc.view_k (o); return o;
}

